// cController.cpp
//
// Copyright 1999 by Benjamin Pickhardt (hackCraft.de).
//

#include "cController.h"

#include "cObject.h"
#include "cWorld.h"

#include <cstdlib>
#include <cassert>
#include <stack>

using namespace std;

#define debug !true
#define debug_transitions true

//--------------------------------

cController::cController(cObject* entity, bool enabled) {
    mDevice = entity;
    mEnabled = enabled;
}


void cController::clearAllInstructions() {
    if (debug) printf("clearInstruction()\n");
    while (!mInstructions.empty()) oldInstruction();
}


void cController::newInstruction(int opcode, tSerial entity, float* vec, int n) {
    if (debug) printf("newInstruction(%s)\n", getInstructionName(opcode));
    //if (entity == 0) return;
    rInstruction* st = new cController::rInstruction;
    st->opcode = opcode;
    st->entity = entity;
    if (vec != NULL) memcpy(st->v, vec, sizeof(float)*3);
    else memset(st->v, 0, sizeof(float)*3);
    st->n = n;
    mInstructions.push(st);
}


void cController::oldInstruction() {
    if (debug) printf("oldInstruction()\n");
    if (!mInstructions.empty()) {
        rInstruction* st = mInstructions.top();
        mInstructions.pop();
        delete st;
    }
}


void cController::process() {
    if (debug) printf("cController::process()\n");

    if (!mEnabled) return;
    if (mDevice == NULL) return;

    if (mInstructions.empty()) {
        waitEnemy();
        return;
    }

    rInstruction* st = mInstructions.top();

    if (debug) printf("Device: %s\n", mDevice->mName.c_str());

    tSerial e = st->entity;
    switch (st->opcode) {
        case WAIT: waitEnemy();
            break;
        case ENGAGE: engageEnemy(e);
            break;
        case TARGET: targetEnemy(e);
            break;
        case ATTACK: attackEnemy(e);
            break;
        case FOLLOW: followLeader(e);
            break;
        case FLEE: fleeEnemy(e);
            break;
        case GOTO: gotoDestination(st->v);
            break;
        case REPEAT: repeatInstructions(st->n);
            break;
        default: printf("Invalid Instruction Request!\n");
            break;
    }
}


void cController::waitEnemy() {
    // Scan world for any enemy in range.
    cObject* nearby = mDevice->enemyNearby();
    if (nearby) {
        if (debug_transitions) {
            printf("> %s Targets %s\n", mDevice->mName.c_str(), nearby->mName.c_str());
        }
        newInstruction(TARGET, nearby->mSerial);
    }

    {
        mDevice->do_idle();
    }
}


void cController::engageEnemy(tSerial entity) {
    {
        mDevice->do_aimFor(entity);
        mDevice->do_moveTowards();
        mDevice->do_aimAt();
    }

    if (mDevice->inTargetRange() > 0.9) {
        if (debug_transitions) {
            cObject* ent = cWorld::instance->mIndex[entity];
            printf("> %s Targets %s\n", mDevice->mName.c_str(), ent->mName.c_str());
        }
        newInstruction(TARGET, entity);
        return;
    }
}


void cController::targetEnemy(tSerial entity) {
    {
        mDevice->do_aimFor(entity);
        mDevice->do_moveTowards();
        mDevice->do_aimAt();
        mDevice->do_fireAt();
    }

    cObject* nearby = mDevice->enemyNearby();
    if (nearby == NULL) {
        oldInstruction();
        return;
    }

    if (mDevice->inMeeleRange() > 0.9) {
        if (debug_transitions) {
            cObject* ent = cWorld::instance->mIndex[entity];
            printf("> %s Attacks %s\n", mDevice->mName.c_str(), ent->mName.c_str());
        }
        newInstruction(ATTACK, nearby->mSerial);
        return;
    }

    if (mDevice->inTargetRange() < 0.1) {
        //printf("inTargetRange() = %.2f\n", inTargetRange());
        if (debug_transitions) {
            cObject* ent = cWorld::instance->mIndex[entity];
            printf("> %s  can't Target %s\n", mDevice->mName.c_str(), ent->mName.c_str());
        }
        oldInstruction(); // Back to whatever was before (possibly engage or wait).
        return;
    }
}


void cController::attackEnemy(tSerial entity) {
    {
        mDevice->do_aimFor(entity);
        mDevice->do_moveNear();
        mDevice->do_aimAt();
        mDevice->do_fireAt();
        //((cMech*)mDevice)->Pattern(tf, "nrnlln");
    }
    
    cObject* nearby = mDevice->enemyNearby();
    if (nearby == NULL) {
        oldInstruction();
        return;
    }

    if (mDevice->inMeeleRange() < 0.1) {
        if (debug_transitions) {
            cObject* ent = cWorld::instance->mIndex[entity];
            printf("> %s can't Attack %s\n", mDevice->mName.c_str(), ent->mName.c_str());
        }
        oldInstruction(); // back to target
        return;
    }
}


void cController::followLeader(tSerial entity) {
    {
        mDevice->do_aimFor(entity);
        mDevice->do_moveNear();
    }
    
    cObject* nearby = mDevice->enemyNearby();
    if (nearby) {
        if (debug_transitions) {
            printf("> %s Targets %s\n", mDevice->mName.c_str(), nearby->mName.c_str());
        }
        newInstruction(TARGET, nearby->mSerial);
        return;
    }
}


void cController::fleeEnemy(tSerial entity) {
}


void cController::gotoDestination(float* v) {
    {
        if (debug) printf("going to: %.2f %.2f %.2f\n", v[0],v[1],v[2]);
        mDevice->do_aimFor(v);
        mDevice->do_moveNear();
    }
    
    if (mDevice->inDestinationRange() > 0.05f) {
        if (debug_transitions) {
            printf("> %s has reached destination\n", mDevice->mName.c_str());
        }
        oldInstruction();
        return;
    }
    
    cObject* nearby = mDevice->enemyNearby();
    if (nearby) {
        if (debug_transitions) {
            printf("> %s Targets %s\n", mDevice->mName.c_str(), nearby->mName.c_str());
        }
        newInstruction(TARGET, nearby->mSerial);
        return;
    }
}


void cController::repeatInstructions(int n) {
    if (n > mInstructions.size() + 1) return;
    printf("################ REPEATING %i #################\n", n);
    stack<rInstruction*> a;
    stack<rInstruction*> b;
    rInstruction* repeat = mInstructions.top();
    mInstructions.pop();
    loopi(n) {
        a.push(mInstructions.top());
        b.push(mInstructions.top());
        mInstructions.pop();
    }
    loopi(n) {
        mInstructions.push(a.top());
        a.pop();
    }
    mInstructions.push(repeat);
    loopi(n) {
        rInstruction* old = b.top();
        b.pop();
        rInstruction* copy = new rInstruction();
        copy->opcode = old->opcode;
        copy->entity = old->entity;
        memcpy(copy->v, old->v, sizeof(float)*3);
        copy->n = old->n;
        mInstructions.push(copy);
    }
}

